//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of NVIDIA CORPORATION nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ''AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Copyright (c) 2008-2021 NVIDIA Corporation. All rights reserved.
// Copyright (c) 2004-2008 AGEIA Technologies, Inc. All rights reserved.
// Copyright (c) 2001-2004 NovodeX AG. All rights reserved.

// This code is auto-generated by the PhysX Clang metadata generator.  Do not edit or be
// prepared for your edits to be quietly ignored next time the clang metadata generator is
// run.  You can find the most recent version of clang metadata generator by contacting
// Chris Nuernberger <chrisn@nvidia.com> or Dilip or Adam.
// The source code for the generate was at one time checked into:
// physx/PhysXMetaDataGenerator/llvm/tools/clang/lib/Frontend/PhysXMetaDataAction.cpp
#define THERE_IS_NO_INCLUDE_GUARD_HERE_FOR_A_REASON

PxVehicleChassisData_PropertiesStart,
PxVehicleChassisData_MMOI,
PxVehicleChassisData_MMass,
PxVehicleChassisData_MCMOffset,
PxVehicleChassisData_PropertiesStop,
PxVehicleEngineData_PropertiesStart,
PxVehicleEngineData_RecipMOI,
PxVehicleEngineData_RecipMaxOmega,
PxVehicleEngineData_MTorqueCurve,
PxVehicleEngineData_MMOI,
PxVehicleEngineData_MPeakTorque,
PxVehicleEngineData_MMaxOmega,
PxVehicleEngineData_MDampingRateFullThrottle,
PxVehicleEngineData_MDampingRateZeroThrottleClutchEngaged,
PxVehicleEngineData_MDampingRateZeroThrottleClutchDisengaged,
PxVehicleEngineData_PropertiesStop,
PxVehicleGearsData_PropertiesStart,
PxVehicleGearsData_GearRatio,
PxVehicleGearsData_MFinalRatio,
PxVehicleGearsData_MNbRatios,
PxVehicleGearsData_MSwitchTime,
PxVehicleGearsData_PropertiesStop,
PxVehicleAutoBoxData_PropertiesStart,
PxVehicleAutoBoxData_Latency,
PxVehicleAutoBoxData_UpRatios,
PxVehicleAutoBoxData_DownRatios,
PxVehicleAutoBoxData_PropertiesStop,
PxVehicleDifferential4WData_PropertiesStart,
PxVehicleDifferential4WData_MFrontRearSplit,
PxVehicleDifferential4WData_MFrontLeftRightSplit,
PxVehicleDifferential4WData_MRearLeftRightSplit,
PxVehicleDifferential4WData_MCentreBias,
PxVehicleDifferential4WData_MFrontBias,
PxVehicleDifferential4WData_MRearBias,
PxVehicleDifferential4WData_MType,
PxVehicleDifferential4WData_PropertiesStop,
PxVehicleDifferentialNWData_PropertiesStart,
PxVehicleDifferentialNWData_DrivenWheelStatus,
PxVehicleDifferentialNWData_PropertiesStop,
PxVehicleAckermannGeometryData_PropertiesStart,
PxVehicleAckermannGeometryData_MAccuracy,
PxVehicleAckermannGeometryData_MFrontWidth,
PxVehicleAckermannGeometryData_MRearWidth,
PxVehicleAckermannGeometryData_MAxleSeparation,
PxVehicleAckermannGeometryData_PropertiesStop,
PxVehicleClutchData_PropertiesStart,
PxVehicleClutchData_MStrength,
PxVehicleClutchData_MAccuracyMode,
PxVehicleClutchData_MEstimateIterations,
PxVehicleClutchData_PropertiesStop,
PxVehicleTireLoadFilterData_PropertiesStart,
PxVehicleTireLoadFilterData_Denominator,
PxVehicleTireLoadFilterData_MMinNormalisedLoad,
PxVehicleTireLoadFilterData_MMinFilteredNormalisedLoad,
PxVehicleTireLoadFilterData_MMaxNormalisedLoad,
PxVehicleTireLoadFilterData_MMaxFilteredNormalisedLoad,
PxVehicleTireLoadFilterData_PropertiesStop,
PxVehicleWheelData_PropertiesStart,
PxVehicleWheelData_RecipRadius,
PxVehicleWheelData_RecipMOI,
PxVehicleWheelData_MRadius,
PxVehicleWheelData_MWidth,
PxVehicleWheelData_MMass,
PxVehicleWheelData_MMOI,
PxVehicleWheelData_MDampingRate,
PxVehicleWheelData_MMaxBrakeTorque,
PxVehicleWheelData_MMaxHandBrakeTorque,
PxVehicleWheelData_MMaxSteer,
PxVehicleWheelData_MToeAngle,
PxVehicleWheelData_PropertiesStop,
PxVehicleSuspensionData_PropertiesStart,
PxVehicleSuspensionData_RecipMaxCompression,
PxVehicleSuspensionData_RecipMaxDroop,
PxVehicleSuspensionData_MassAndPreserveNaturalFrequency,
PxVehicleSuspensionData_MSpringStrength,
PxVehicleSuspensionData_MSpringDamperRate,
PxVehicleSuspensionData_MMaxCompression,
PxVehicleSuspensionData_MMaxDroop,
PxVehicleSuspensionData_MSprungMass,
PxVehicleSuspensionData_MCamberAtRest,
PxVehicleSuspensionData_MCamberAtMaxCompression,
PxVehicleSuspensionData_MCamberAtMaxDroop,
PxVehicleSuspensionData_PropertiesStop,
PxVehicleAntiRollBarData_PropertiesStart,
PxVehicleAntiRollBarData_MWheel0,
PxVehicleAntiRollBarData_MWheel1,
PxVehicleAntiRollBarData_MStiffness,
PxVehicleAntiRollBarData_PropertiesStop,
PxVehicleTireData_PropertiesStart,
PxVehicleTireData_RecipLongitudinalStiffnessPerUnitGravity,
PxVehicleTireData_FrictionVsSlipGraphRecipx1Minusx0,
PxVehicleTireData_FrictionVsSlipGraphRecipx2Minusx1,
PxVehicleTireData_MLatStiffX,
PxVehicleTireData_MLatStiffY,
PxVehicleTireData_MLongitudinalStiffnessPerUnitGravity,
PxVehicleTireData_MCamberStiffnessPerUnitGravity,
PxVehicleTireData_MType,
PxVehicleTireData_MFrictionVsSlipGraph,
PxVehicleTireData_PropertiesStop,
PxVehicleWheels4SimData_PropertiesStart,
PxVehicleWheels4SimData_TireRestLoadsArray,
PxVehicleWheels4SimData_RecipTireRestLoadsArray,
PxVehicleWheels4SimData_PropertiesStop,
PxVehicleWheelsSimData_PropertiesStart,
PxVehicleWheelsSimData_ChassisMass,
PxVehicleWheelsSimData_SuspensionData,
PxVehicleWheelsSimData_WheelData,
PxVehicleWheelsSimData_TireData,
PxVehicleWheelsSimData_SuspTravelDirection,
PxVehicleWheelsSimData_SuspForceAppPointOffset,
PxVehicleWheelsSimData_TireForceAppPointOffset,
PxVehicleWheelsSimData_WheelCentreOffset,
PxVehicleWheelsSimData_WheelShapeMapping,
PxVehicleWheelsSimData_SceneQueryFilterData,
PxVehicleWheelsSimData_AntiRollBarData,
PxVehicleWheelsSimData_TireLoadFilterData,
PxVehicleWheelsSimData_MinLongSlipDenominator,
PxVehicleWheelsSimData_Flags,
PxVehicleWheelsSimData_ThresholdLongSpeed,
PxVehicleWheelsSimData_LowForwardSpeedSubStepCount,
PxVehicleWheelsSimData_HighForwardSpeedSubStepCount,
PxVehicleWheelsSimData_WheelEnabledState,
PxVehicleWheelsSimData_PropertiesStop,
PxVehicleWheelsDynData_PropertiesStart,
PxVehicleWheelsDynData_TireForceShaderFunction,
PxVehicleWheelsDynData_WheelRotationSpeed,
PxVehicleWheelsDynData_WheelRotationAngle,
PxVehicleWheelsDynData_Wheel4DynData,
PxVehicleWheelsDynData_Constraints,
PxVehicleWheelsDynData_PropertiesStop,
PxVehicleWheels_PropertiesStart,
PxVehicleWheels_VehicleType,
PxVehicleWheels_RigidDynamicActor,
PxVehicleWheels_ConcreteTypeName,
PxVehicleWheels_MWheelsSimData,
PxVehicleWheels_MWheelsDynData,
PxVehicleWheels_PropertiesStop,
PxVehicleDriveDynData_PropertiesStart,
PxVehicleDriveDynData_AnalogInput,
PxVehicleDriveDynData_GearUp,
PxVehicleDriveDynData_GearDown,
PxVehicleDriveDynData_UseAutoGears,
PxVehicleDriveDynData_CurrentGear,
PxVehicleDriveDynData_TargetGear,
PxVehicleDriveDynData_EngineRotationSpeed,
PxVehicleDriveDynData_GearChange,
PxVehicleDriveDynData_GearSwitchTime,
PxVehicleDriveDynData_AutoBoxSwitchTime,
PxVehicleDriveDynData_MUseAutoGears,
PxVehicleDriveDynData_MGearUpPressed,
PxVehicleDriveDynData_MGearDownPressed,
PxVehicleDriveDynData_MCurrentGear,
PxVehicleDriveDynData_MTargetGear,
PxVehicleDriveDynData_MEnginespeed,
PxVehicleDriveDynData_MGearSwitchTime,
PxVehicleDriveDynData_MAutoBoxSwitchTime,
PxVehicleDriveDynData_PropertiesStop,
PxVehicleDriveSimData_PropertiesStart,
PxVehicleDriveSimData_EngineData,
PxVehicleDriveSimData_GearsData,
PxVehicleDriveSimData_ClutchData,
PxVehicleDriveSimData_AutoBoxData,
PxVehicleDriveSimData_PropertiesStop,
PxVehicleDriveSimData4W_PropertiesStart,
PxVehicleDriveSimData4W_DiffData,
PxVehicleDriveSimData4W_AckermannGeometryData,
PxVehicleDriveSimData4W_PropertiesStop,
PxVehicleDrive_PropertiesStart,
PxVehicleDrive_ConcreteTypeName,
PxVehicleDrive_MDriveDynData,
PxVehicleDrive_PropertiesStop,
PxVehicleDrive4W_PropertiesStart,
PxVehicleDrive4W_ConcreteTypeName,
PxVehicleDrive4W_MDriveSimData,
PxVehicleDrive4W_PropertiesStop,
PxVehicleDriveTank_PropertiesStart,
PxVehicleDriveTank_DriveModel,
PxVehicleDriveTank_ConcreteTypeName,
PxVehicleDriveTank_MDriveSimData,
PxVehicleDriveTank_PropertiesStop,
PxVehicleDriveSimDataNW_PropertiesStart,
PxVehicleDriveSimDataNW_DiffData,
PxVehicleDriveSimDataNW_PropertiesStop,
PxVehicleDriveNW_PropertiesStart,
PxVehicleDriveNW_ConcreteTypeName,
PxVehicleDriveNW_MDriveSimData,
PxVehicleDriveNW_PropertiesStop,
PxVehicleNoDrive_PropertiesStart,
PxVehicleNoDrive_BrakeTorque,
PxVehicleNoDrive_DriveTorque,
PxVehicleNoDrive_SteerAngle,
PxVehicleNoDrive_ConcreteTypeName,
PxVehicleNoDrive_PropertiesStop,


#undef THERE_IS_NO_INCLUDE_GUARD_HERE_FOR_A_REASON
